;;;;;;;;;;;;;;;;;;;;;;;;;;; MOT-EW Model AH V5 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; This model completes a multiple object tracking task where there are  
;;; four quadrants, N friendlies, and N hostiles. The goal is to turn on
;;; an alarm when a hostile enters a quadrant and off when it exits. The
;;; friendlies serve as distractors. An additional condition includes
;;; electronic warfare where a hostile disappears for N seconds. When 
;;; ew occurs, the model turns on an ew alarm not specific to a quadrant.

;;; NOTE: the experiment code is implemented separately in Python to 
;;; generate the agi and update the visicon. 

;;; Authors      	: Alexander Hough and Christopher Stevens
;;; Emails			: Alexander.hough.1@us.af.mil
;;;					: Christopher.Stevens.28@us.af.mil
;;; Affiliation     : Cognitive Models Branch (RHWM)
;;;             	: Air Force Research Laboratory
;;;             	: Wright Patterson AFB, OH, 45433

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Model Notes and TODOs
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; V2 modifications

	; 1. Ticks and EW checks
	;	Modified model so that ticks are started wtih start production and are reset during ew check. 
		; Time since ew check updating was removed from find-object production. 
		; The temporal module is cleared when the ew check is initiated and started after an ew decision
		; Ticks now only serve the purpose of tracking time since last ew check 
		; All ew time slots and their references were removed 
		
	; 2. The productions to handle disapparing hostiles during ew (ew-decision-hostile-vanish) have been modified 
		; They were only able to fire during ew checks, however a hostile can disappear at any time during DUAL task
		; They were changed to be able to fire at any time during the DUAL task if a hostile disappears 
		; Required adding ew-alarm slot in goal buffer to ensure the ew alarm is turned on or off appropriately when
		;	not doing an ew check 
		
	; 3. Added expected-hostiles slot in goal buffer
		; This keeps the highest hostile count (total-hostiles) during an ew check. It represents the
		; 	expected amount of hostiles to guide ew checks rather than use hard coding.
		
	; 4. Modified all ew decision productions to use expected-hostiles
		; Productions now use expected-hostiles to compare with hostile count to determine if there are more, less, or 
		; 	the amount of hostiles expected to make decisions. In case an ew attack happens at task start, the conditions
		; 	for finding the expected amount is not >= if the expected amount is less than the actual (ew attack at start). 
		; Added a conditional to encode-hostile to set expected hostile when doing ew check 
		
	; 5. Added wait-for-key-press production to force the model to wait until any alarm key press has been initiated before
		; before checking the next quadrants alarm state
		; This should eliminate cases where the model tried to turn an alarm on or off twice, then corrected itself.  
			; This can result in the model turning on, then off, then back on alarms or vice-versa. 
			
	; 6. Removed wait-for-key-press and added manaual buffer check to check alarm production.
		; Preparation is busy for less time (about 75 ms wait - three productions firing at 25 ms) than processor (about 225 ms)
			
; V3 Modifications - trying to speed up model response times for faster moving objects 
	; 1. Force model to search quads in clockwise fashion - the model may search NE quad first, then last in next cycle - long time
		; Only needed to adjust quad slot search values in quad search productions (e.g., for NE: NW yes NE no SE no SW no)
		
	; 2. Sped up two motor movement parameters to enhance alarm activation respond time (see below)
			;:motor-feature-prep-time .025 ; (default = .05) The time in seconds that it takes to prepare each movement feature. 
			;:motor-initiation-time .025 ; (default = .05) This is the length of the initiation time for motor actions in seconds
			
	; 3. Reduce the need to attend and encode every single object
		; Only find friendlies, but don't shift attention and encode
		
; V4 Modifications
	; 1. During MOT, the model quits searching a quadrant when a hostile is found rather than doing complete search
		; Hostile is attended and a decision is made - alarm on productions modified and changed name to hostile-found-...
			; Modifications for alarm-on productions include condition that a hostile is currently attended
			; Modifications for alarm-off productions include clearing the visual buffer. 
		; Encode-hostile production that counts hostiles is now exclusive to EW checks. 
		
	; 2. Was going to move the alarm check to after encode to reduce the need to check the manual buffer state to reduce
	;    errors when the model checks an alarm before a key-press completes to turn on/off an alarm
		; Not needed when parameters are at defaults because processing takes longer allowing the key-press and alarm setting
			; to occur before check alarm. If parameters like production fire time and attention shift speed are lowered, this
			; will become a problem and the manual buffer state check needs to be included or check alarm moved. 
		; Likely not needed as model now searched quads in clockwise fashion so enough time should pass between between checking
			; the same quad again. 
			
; V5 Modifications - bug where model finds hostile, but fails to attend to it as it has moved too far out of visual location area
	; 1. Added two productions to handle rare edge cases where the model doesn't find an object after shifting attention or for 
		; some reason a friendly is found when attention was shifted to a hostile. 
		; Production 1 (found-hostile-but-moved-too-far) handles when an object is not in the location attention shifted to
		; Production 2 (found-hostile-attended-friendly) handles cases where a friendly is found when shifting attention to a hostile
	; 2. Switch from key press response to punch
		; This reduces key presses by at least 50ms and assumes hands are on the home keys
		; The mapping from NW clockwise to SW is: f j k d. The key for EW is the space bar (spc)
		; No alarm on and off productions modified to use punch instead of key press and goal chunk has hand and finger slots instead
			; of just the key slot

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; MODEL DEFINITION ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(clear-all)

(require-extra "fatigue")

(define-model mot-ew-model

	;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	; Start model parameters
	;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    ;; set global parameters
    (sgp
    	;; Standard parameters 
        :v t ;Print trace in cmd window (t for yes and nil for no)
        :trace-detail low ;Trace detail low, medium, or high
		:trace-filter production-firing-only ;Control which aspects of trace are output
        :show-focus t ; If the task window is showing, this controls whether you can see where the model's attention is (red circle)
        :esc t ;Enable subsymbolic features
        :cmdt nil ;Command trace
		
        ;; visual finsts
        :visual-num-finsts 16 ;# visual items marked as attended (default = 4); increased to max 3 of objects in condition
        :visual-finst-span 3 ;Seconds visual items remain marked as attended (default = 3)
		
		;; Visual attention 
		:visual-attention-latency .085 ;Speed of visual attention shift (Default is .085)	
		:visual-movement-tolerance 1.5 ; how far an object can move and still considered the same object 
									  ; without being tracked. It is measured in degrees of visual angle
									  ; default is .5
		
		;; Production parameters - Utility learning and noise have no effect in the current model as there is no competition between 
			;; productions and partial matching is turned off.
		:ul t ;Utility learning on/off
		:egs 1 ; noise parameter (production utility) - adds random noise to production utilities.
		:dat .05 ; Production fire time (default .05) - controls how long it takes for productions to fire
		
		; Manual buffer - key presses 
		:motor-feature-prep-time .05 ; (default = .05) The time in seconds that it takes to prepare each movement feature. 
		:motor-initiation-time .05 ; (default = .05) This is the length of the initiation time for motor actions in seconds
		
    )

	(chunk-type goal task state quad last-quad alarm hand finger left right top bottom NWquad NEquad SEquad SWquad Nhostiles expected-hostiles ew-check ew-alarm) ; Goal chunk (executive control/WM)

	(add-dm
		; Basic state chunks
		(start isa chunk) (attend isa chunk) (find-object isa chunk) 
		(ID-object isa chunk) (search-quad isa chunk) (check-quad isa chunk)
		(check-alarm isa chunk)
		(MOT isa chunk) (EW isa chunk) (DUAL isa chunk)
		(NW isa chunk) (NE isa chunk) (SE isa chunk) (SW isa chunk)
		; ew stuff
		(ew-search-alarm) (ew-all isa chunk) 
		(yes isa chunk) (no isa chunk)
		; Goal chunk
		(goal1 isa goal state start NWquad no NEquad no	SEquad no SWquad no)
	)
		(goal-focus goal1)
	
	;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	; End model parameters
	;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; MOT and general use productions 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(p start ; Starts the model - mainly to set the temporal module
	=goal>
		state					start
		task					=task
		left					nil	; These are coordinates for quadrants and allow the model to do quad search
		right		            nil
		top			            nil
		bottom             		nil
==>
	=goal>
		ew-check				no	; Whether an ew check is happening (yes or no)
		Nhostiles				0 	; For counting hostiles
		expected-hostiles		0   ; For keeping track of how many hostiles are in given condition (equal to highest total-hostiles)
		alarm					black ; Assume alarms are off starting
		ew-alarm				black ; Assume ew alarm is off starting
	+temporal>
        isa                     time			; Start tick timer
        ticks                   0
	!eval!						(cond ; Set state to check ew alarm if task is ew, if not set to search quad for MOT/DUAL
								((chunk-slot-equal =task 'EW) (mod-buffer-chunk 'goal '(state ew-search-alarm)))
								((or (chunk-slot-equal =task 'MOT) (chunk-slot-equal =task 'DUAL)) (mod-buffer-chunk 'goal '(state search-quad)))
								)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; START - Quad search productions - orient model to search a given quad ;;;

(p search-NW-quad ; Initiate search of NW quad by looking at the quad and finding the location of the alarm line(s) if not doing EW check
	=goal>
		state					search-quad
		NWquad					no ; This quad has not been searched
		NEquad					no ; Other quads not yet searched (forcing this one first)
		SWquad					no ;
		SEquad					no ; 
		ew-check				no
		- last-quad				NW ; This quad was not just searched in the last completed quad search loop
	?visual>
		state					free
==>
	=goal> 
		state					check-alarm
		quad					NW
		hand					left
		finger					index
		left					300 ; Coordinates of quadrant 
		right		            600
		top			            300
		bottom             		600
	+visual-location>							
		:attended				nil
		kind					line
		= height				298 ; Vertical alarm indicator line
		= width					0
		>= screen-x             300 ; Coordinates of quadrant 
		<= screen-x             600
		>= screen-y             300
		<= screen-y             600
)

(p search-NE-quad ; Initiate search of NE quad by looking at the quad and finding the location of the alarm line(s) if not doing EW check
	=goal>
		state					search-quad
		NEquad					no ; This quad has not been searched
		NWquad					yes ; Just searched
		SWquad					no ; Other quads not yet searched (forcing this one second)
		SEquad					no ; 
		ew-check				no
		- last-quad				NE ; This quad was not just searched in the last completed quad search loop
	?visual>
		state					free
==>
	=goal>
		state					check-alarm
		quad					NE
		hand					right
		finger					index
		left					600 ; Coordinates of quadrant 
		right		            900
		top			            300
		bottom             		600
	+visual-location>							
		:attended				nil
		kind					line
		= height				298 ; Vertical alarm indicator line
		= width					0
		>= screen-x             600 ; Coordinates of quadrant 
		<= screen-x             900
		>= screen-y             300
		<= screen-y             600
)

(p search-SE-quad ; Initiate search of SE quad by looking at the quad and finding the location of the alarm line(s) if not doing EW check
	=goal>
		state					search-quad
		SEquad					no ; This quad has not been searched		
		NWquad					yes ; Just searched
		NEquad					yes ; Just searched
		SWquad					no ;  SW not yet searched (forcing this one third)
		ew-check				no
		- last-quad				SE ; This quad was not just searched in the last completed quad search loop
	?visual>
		state					free
==>
	=goal>
		state					check-alarm
		quad					SE
		hand					right
		finger					middle
		left					600 ; Coordinates of quadrant 
		right		            900
		top			            600
		bottom             		900
	+visual-location>							
		:attended				nil
		kind					line
		= height				298 ; Vertical alarm indicator line
		= width					0
		>= screen-x             600 ; Coordinates of quadrant 
		<= screen-x             900
		>= screen-y             600
		<= screen-y             900
)

(p search-SW-quad ; Initiate search of SW quad by looking at the quad and finding the location of the alarm line(s) if not doing EW check
	=goal>
		state					search-quad
		SWquad					no ;  This quad has not been searched
		SEquad					yes ; Just searched
		NWquad					yes ; Just searched
		NEquad					yes ; Just searched		
		ew-check				no
		- last-quad				SW ; This quad was not just searched in the last completed quad search loop
	?visual>
		state					free
==>
	=goal>
		state					check-alarm
		quad					SW
		hand					left
		finger					middle
		left					300 ; Coordinates of quadrant 
		right		            600
		top			            600
		bottom             		900
	+visual-location>							
		:attended				nil
		kind					line
		= height				298 ; Vertical alarm indicator line
		= width					0
		>= screen-x             300 ; Coordinates of quadrant 
		<= screen-x             600
		>= screen-y             600
		<= screen-y             900
)

;;; END - Quad search productions - orient model to search a given quad ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; START - check color of quad or ew alarm indicator line ;;;

(p check-alarm ; Check if alarm is on or off using "peripheral vision" and set alarm slot in goal buffer
	=goal>
		state					check-alarm
		- quad					nil
		- left					nil
		- right		            nil
		- top			        nil
		- bottom             	nil
		ew-check				=ew-check 
	=visual-location>
		kind					line
		color					=color 	; Green is alarm-on and black is alarm-off
	; ?manual>							; Don't check alarm until any alarm key-press has been initiated, so that it doesn't check alarm start before alarm is correctly turned on/off
		; ;Processor 				free ;Processor busy when request is received and after preparation of movement
		; Preparation 			free ;preparation is busy when request is received 
==>
	=goal>
		state					find-object
		;alarm					=color	; set alarm state to color of alarm indicator 
	; Added conditional to set either the alarm or ew-alarm depending on whether an ew check is occuring 
		; needed to separate alarms for rare cases were a hostile disappears while the model is moving 
		; attention and is handled by the ew-decision-hostile-vanish productions. 
	!eval!						(cond
								((chunk-slot-equal =ew-check 'yes) (mod-buffer-chunk 'goal '(ew-alarm =color)))
								((chunk-slot-equal =ew-check 'no) (mod-buffer-chunk 'goal '(alarm =color)))
								)
)

;;; END - check color of quad or ew alarm indicator line ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; START - Find-attend-encode loop ;;;

(p find-object ; Find an object not yet attended in the quad currently searching and update time since last EW check if task is EW or DUAL
	=goal>
		state					find-object
		task					=task
		left					=left
		right		            =right
		top			            =top
		bottom             		=bottom
	?visual>
		state					free
==>
	+visual-location>							
		:nearest				clockwise 	; Search in clockwise fashion
		:attended				nil			; Not marked as attended
		kind					text		
		- value					"ew"		; not "ew" as that is supposed to be invisible
		>= screen-x             =left		; Within coordinates representing quadrant
		<= screen-x             =right
		>= screen-y             =top
		<= screen-y             =bottom
	=goal>
		state					attend		
)

(p attend-object ; Move attention to found object
	=goal>
		state					attend
	=visual-location>
		kind					text
		- color					red
	?visual>
		state					free
==>
	+visual>
		cmd 					move-attention
		screen-pos				=visual-location
	=goal>
		state					ID-object
)

(p found-friendly-no-attend ; ID attended object as friendly 
	=goal>
		state					attend
		quad					=quad
	=visual-location>		
		kind					text		; Identify object as F if is text and is red
		color					red
	?visual>
		state					free
==>
	=goal>
		state					find-object
	+visual>
		cmd						assign-finst 
		location 				=visual-location
	!output!					(Friendly found in =quad)
)

	;;; START - bug fix ;;;
	; This is a bug fix for an object moving out of the found location prior to attention shift
(p found-hostile-but-moved-too-far ; If a hostile was found, but the object was not in that location when attention shifted, handle and find another object
	=goal>
		state					ID-object
		quad					=quad
	=visual-location>
	?visual>
		buffer 					failure
==>
	=goal>
		state					find-object
	!eval!						(remove-visual-finsts t nil)
	!output!					(Hostile found but moved too far from location to attend in =quad)
)
	; This is a bug fix where a friendly is found when attention was shifted towards a found hostile
(p found-hostile-attended-friendly ; If a hostile was found, but the object was not in that location when attention shifted, handle and find another object
	=goal>
		state					ID-object
		quad					=quad
	=visual>
		value					"F"
==>
	=goal>
		state					find-object
	+visual>
		cmd						clear
	!output!					(Hostile found but friendly attended in =quad)
)
	;;; END - bug fix ;;;

;;; END - Find-attend-encode loop ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; START - alarm on/off decisions ;;;

(p hostile-found-alarm-on-ok ; All objects searched, the alarm is on and there are hostiles...do nothing and move to next quad
	=goal>
		state					ID-object
		quad					=quad
		alarm					green 	; Alarm is on
		ew-check				no		; not doing ew check
		- quad 					ew-all	; not doing ew check
	=visual>
		value					"H"
		color					=color	
==>
	=goal>
		last-quad				=quad
		state					search-quad
		hand					nil
		finger					nil
	+visual>
		cmd						clear
	!eval!						(remove-visual-finsts t nil)
	; This conditional marks the current quad being searched as completed (places yes chunk in respective slot),...
	; sets the state chunk to go back to quad search, and if currently doing an EW check, it sets the state to ew-decision
	!eval!						(cond
									((chunk-slot-equal =quad 'NW) (mod-buffer-chunk 'goal '(NWquad yes)))
									((chunk-slot-equal =quad 'NE) (mod-buffer-chunk 'goal '(NEquad yes)))
									((chunk-slot-equal =quad 'SE) (mod-buffer-chunk 'goal '(SEquad yes)))
									((chunk-slot-equal =quad 'SW) (mod-buffer-chunk 'goal '(SWquad yes)))
								)	
	!output!					(=quad searched... hostile found and alarm on... moving to next quad)
)

(p hostile-found-turn-on-alarm ; All objects searched, the alarm is off and there are hostiles...turn on the alarm then move to next quad
	=goal>
		state					ID-object
		quad					=quad
		alarm					black	; alarm is off
		ew-check				no		; not doing ew check
		- quad 					ew-all	; not doing ew check
		hand					=hand
		finger					=finger
	=visual>
		value					"H"
		color					=color
	?manual>
		state					free		
==>
	+manual> ; Turn on alarm by pressing quad alarm key
		cmd						punch
		hand					=hand
		finger					=finger
	=goal>
		state					search-quad
		last-quad				=quad
		hand					nil
		finger					nil
	+visual>
		cmd						clear
	; This conditional marks the current quad being searched as completed (places yes chunk in respective slot)
	!eval!						(cond
									((chunk-slot-equal =quad 'NW) (mod-buffer-chunk 'goal '(NWquad yes)))
									((chunk-slot-equal =quad 'NE) (mod-buffer-chunk 'goal '(NEquad yes)))
									((chunk-slot-equal =quad 'SE) (mod-buffer-chunk 'goal '(SEquad yes)))
									((chunk-slot-equal =quad 'SW) (mod-buffer-chunk 'goal '(SWquad yes)))
								)
	!eval!						(remove-visual-finsts t nil)
	!output!					(=quad searched... hostile found and alarm off... turning on alarm and moving to next quad)
)

(p all-objects-searched-alarm-off-ok ; All objects searched, the alarm is off and there are no hostiles...do nothing and move to next quad
	=goal>
		state					attend
		quad					=quad
		alarm					black 	; Alarm is off
		ew-check				no		; Not doing ew check		
		- quad 					ew-all  ; Not doing ew check
	?visual>
		state       			free
	?visual-location>
		state					error	
==>
	=goal>
		last-quad				=quad
		state					search-quad
		Nhostiles				0
		hand					nil
		finger					nil
	!eval!						(remove-visual-finsts t nil)
	; This conditional marks the current quad being searched as completed (places yes chunk in respective slot),...
	; sets the state chunk to go back to quad search, and if currently doing an EW check, it sets the state to ew-decision
	!eval!						(cond
									((chunk-slot-equal =quad 'NW) (mod-buffer-chunk 'goal '(NWquad yes)))
									((chunk-slot-equal =quad 'NE) (mod-buffer-chunk 'goal '(NEquad yes)))
									((chunk-slot-equal =quad 'SE) (mod-buffer-chunk 'goal '(SEquad yes)))
									((chunk-slot-equal =quad 'SW) (mod-buffer-chunk 'goal '(SWquad yes)))
								)	
	!output!					(=quad searched... no hostiles found... moving to next quad)
)

(p all-objects-searched-turn-off-alarm ; All objects searched, the alarm is on and there are no hostiles...turn off the alarm then move to next quad
	=goal>
		state					attend
		quad					=quad
		alarm					green	; alarm is on
		ew-check				no		; not doing ew check
		- quad 					ew-all	; not doing ew check
		hand					=hand
		finger					=finger
	?visual>
		state       			free
	?visual-location>
		state					error	
	?manual>
		state					free		
==>
	+manual>	; turn off alarm by pressing quad alarm key
		cmd						punch
		hand					=hand
		finger					=finger
	=goal>
		state					search-quad
		last-quad				=quad
		hand					nil
		finger					nil
		Nhostiles				0
	; This conditional marks the current quad being searched as completed (places yes chunk in respective slot)
	!eval!						(cond
									((chunk-slot-equal =quad 'NW) (mod-buffer-chunk 'goal '(NWquad yes)))
									((chunk-slot-equal =quad 'NE) (mod-buffer-chunk 'goal '(NEquad yes)))
									((chunk-slot-equal =quad 'SE) (mod-buffer-chunk 'goal '(SEquad yes)))
									((chunk-slot-equal =quad 'SW) (mod-buffer-chunk 'goal '(SWquad yes)))
								)
	!eval!						(remove-visual-finsts t nil)
	!output!					(=quad searched... no hostiles found... turning on alarm and moving to next quad)
)

;;; END - alarm on/off decisions ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; START - Quad search loop completed ;;;

; * NOTE *
; This production can be modified to change ew check strategy. It will initiate an ew check in the
; dual condition by uncommenting the eval statement and commenting out the state slot change on the 
; RHS (state search-quad). If doing so, this will initiate an ew check after all quads have been 
; searched. Also need to comment out the ew-initiate-check production below. 

; Ew check after t ticks: 
; Eval is commented out and ew-initiate-check production is not commented out, the model will
; initiate ew checks after t ticks have passed. 

; Ew check after all quads searched: 
; Eval statement is uncommented, the state slot change commented out, and the ew-initiate-check 
; is commented out then the model will do ew checks 

(p all-quads-searched ; All quads have been searched...reset and start searching through quads again 
	=goal>
		state					search-quad
		quad					=quad
		NWquad					yes	; All quads have been searched
		NEquad					yes
		SEquad					yes
		SWquad					yes
		ew-check				no
		task					=task
==>
	=goal>
		; state					search-quad
		NWquad					no	; Clear quad searched slots
		NEquad					no
		SEquad					no
		SWquad					no
		last-quad				=quad ; Set quad last searched - avoid searching the same one immediately after reset 
	!output!					(all quads searched... restart quad search)
	; This eval will initiate an ew check after all quads have been searched, see note above for instructions. 
	!eval!						(cond 
								((chunk-slot-equal =task 'DUAL) (mod-buffer-chunk 'goal '(state ew-search-alarm)))
								((chunk-slot-equal =task 'MOT) (mod-buffer-chunk 'goal '(state search-quad)))
								)
)

;;; END - Quad search loop completed ;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; EW specific productions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 

; (p ew-initiate-check ; If a certain period of time has passed since the last EW check, start an EW check (DUAL condition)
	; =goal>
		; state					search-quad
		; ew-check				no		; Not currently doing ew check 
		; task					DUAL	; Task is dual
		; Nhostiles				0		; Only divert attention to ew check if a quad was just searched or no hostiles were found
	; =temporal>
		; > ticks					25
	; ?visual>
		; state					free							
; ==>
	; =goal>
		; state					ew-search-alarm
; !output!						(Starting EW check)
; )

(p ew-search-alarm ; Check the state of the EW alarm
	=goal>
		state					ew-search-alarm
		ew-check				no		; ew check not ongoing
		Nhostiles				0
	?visual>
		state					free
	=temporal>
==>
	=goal>
		state					check-alarm ;Set to use general productions: check-alarm, find, attend, and encode (see above)
		ew-check				yes			; indicate ew check is ongoing now
		quad					ew-all		; Set quad for ew check
		left					300			; Coordinates includes all quads
		right		            900
		top			            300
		bottom             		900
	+visual-location>							
		:attended				nil
		kind					line
		= height				596			; This is the ew alarm indicator line
		= width					0
		>= screen-x             300
		<= screen-x             600
		>= screen-y             300
		<= screen-y             600
	-temporal>
)

(p ew-encode-hostile ; ID attended object as hostile
	=goal>
		state					ID-object
		quad					=quad
		Nhostiles				=NH ; Keep hostile count
		expected-hostiles		=expected
		ew-check				yes			
	=visual>
		value					"H"
		color					=color	
	!bind!	=new-count			(+ =NH 1) ; Update count
==>
	=goal>
		state					find-object	
		Nhostiles				=new-count ; update
	+visual>
		cmd						clear
	!output!					(Hostile found... count is =new-count)
	; Conditional sets expected number of hostiles as the largest amount counted only in ew condition as it treats all quads as asingle one (ew-all)
	!eval!						(cond
								((> =new-count =expected) (mod-buffer-chunk 'goal '(expected-hostiles =new-count)))
								)
)

(p ew-decision-hostile-vanish-turn-on-alarm ; If an object disappeared (changed to ew)...turn on the alarm and end EW check
	=goal>
		state					ID-object
		quad					=quad
		;ew-check				yes		; EW check ongoing
		- task					MOT
		task					=task
		ew-alarm				black	; EW alarm is off 
	=visual>
		value					"ew"	; The attended object is "ew" which is technically invisible - model saw it disappear
		color					=color	
	?manual>
		state					free
==>
	+manual>	; Turn on ew alarm
		cmd						punch
		hand					right
		finger					thumb
	=goal>
		Nhostiles				0
		ew-alarm				green  ; Set alarm state to green (on) 
	+temporal>
		isa 					time
		ticks					0
	+visual>
		cmd						clear
	!eval!						(remove-visual-finsts t nil)
	; If task is EW, set state to find-object and ew-check to yes - this will bypass checking ew-alarm as its silly to keep checking ew-alarm in EW only task 
	; If task is MOT/DUAL set state to search-quad and ew-check to no - this will initiate quad search 
	!eval!						(cond
								((chunk-slot-equal =task 'EW) (and (mod-buffer-chunk 'goal '(state find-object)) (mod-buffer-chunk 'goal '(ew-check yes))))
								((chunk-slot-equal =task 'MOT) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								((chunk-slot-equal =task 'DUAL) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								)
	!output!					(Hostile disappeared and ew alarm is off...turning on ew alarm)
)

; ; NOTE: This is for a bug in the python code where EW causes more than one hostiles to disappear. Remove after fixing EW bug
(p ew-decision-hostile-vanish-alarm-ok ; If an object disappeared (changed to ew) and the ew alarm is on...do nothing and end EW check
	=goal>
		state					ID-object
		quad					=quad
		;ew-check				yes		; EW check ongoing
		- task					MOT
		task					=task
		ew-alarm				green	; EW alarm is on
	=visual>
		value					"ew"	; The attended object is "ew" which is technically invisible - model saw it disappear
		color					=color	
==>
	=goal>
		Nhostiles				0	
	+temporal>
		isa 					time
		ticks					0
	+visual>
		cmd						clear
	!eval!						(remove-visual-finsts t nil)
	; If task is EW, set state to find-object and ew-check to yes - this will bypass checking ew-alarm as its silly to keep checking ew-alarm in EW only task 
	; If task is MOT/DUAL set state to search-quad and ew-check to no - this will initiate quad search 
	!eval!						(cond
								((chunk-slot-equal =task 'EW) (and (mod-buffer-chunk 'goal '(state find-object)) (mod-buffer-chunk 'goal '(ew-check yes))))
								((chunk-slot-equal =task 'MOT) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								((chunk-slot-equal =task 'DUAL) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								)
)

(p ew-decision-turn-on-alarm ; If the number of hostiles counted is less than expected and the alarm is off...turn on the alarm and end the EW check 
	=goal>
		state					attend
		quad					ew-all
		ew-check				yes				; ew check on going
		ew-alarm				black 			; ew alarm is off
		expected-hostiles		=expected		; expected hostiles based on quad search cycle with highest count
		< Nhostiles				=expected 		; There is less hostile than expected
		Nhostiles				=count
		task					=task
	?manual>
		state					free
	?visual>
		state       			free
	?visual-location>
		state					error
==>
	+manual>	; Turn on ew alarm
		cmd						punch
		hand					right
		finger					thumb
	=goal>		
		Nhostiles				0
		ew-alarm					green ; set ew alarm state to on
	+temporal>
		isa 					time
		ticks					0
	; If task is EW, set state to find-object and ew-check to yes - this will bypass checking ew-alarm as its silly to keep checking ew-alarm in EW only task 
	; If task is MOT/DUAL set state to search-quad and ew-check to no - this will initiate quad search 
	!eval!						(cond
								((chunk-slot-equal =task 'EW) (and (mod-buffer-chunk 'goal '(state find-object)) (mod-buffer-chunk 'goal '(ew-check yes))))
								((chunk-slot-equal =task 'MOT) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								((chunk-slot-equal =task 'DUAL) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								)
	!eval!						(remove-visual-finsts t nil)
	!output!					(=count hostiles were found but =expected were expected and the ew alarm is off...turning on ew alarm)
)

(p ew-decision-turn-off-alarm ; If the number of hostiles counted is the amount expected and the alarm is on..turn off the alarm and end the EW check 
	=goal>
		state					attend
		task					=task
		quad					ew-all			; ew check ongoing
		ew-check				yes				; ew check ongoing
		ew-alarm				green			; ew alarm is on
		expected-hostiles		=expected		; expected hostiles based on quad search cycle with highest count
		>= Nhostiles			=expected 		; The amount of hostiles encoded/counted is expected or greater than expected (if ew occured at start of task)
		Nhostiles				=count
	?manual>
		state					free
	?visual>
		state       			free
	?visual-location>
		state					error
==>
	+manual> ; Turn off the ew alarm
		cmd						punch
		hand					right
		finger					thumb
	=goal>	
		Nhostiles				0
		ew-alarm				black 			; Set ew alarm state to off
	+temporal>
		isa 					time
		ticks					0
	; If task is EW, set state to find-object and ew-check to yes - this will bypass checking ew-alarm as its silly to keep checking ew-alarm in EW only task 
	; If task is MOT/DUAL set state to search-quad and ew-check to no - this will initiate quad search 
	!eval!						(cond
								((chunk-slot-equal =task 'EW) (and (mod-buffer-chunk 'goal '(state find-object)) (mod-buffer-chunk 'goal '(ew-check yes))))
								((chunk-slot-equal =task 'MOT) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								((chunk-slot-equal =task 'DUAL) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								)
	!eval!						(remove-visual-finsts t nil)
	!output!					(=count hostiles were found and =expected were expected... the ew alarm is on...turning off now)
)

(p ew-decision-alarm-on-ok ; If the number of hostiles counted is less than expected and the alarm is on...do nothing and end the EW check 
	=goal>
		state					attend
		quad					ew-all			; ew check ongoing
		ew-check				yes				; ew check ongoing
		ew-alarm				green			; ew alarm is on
		expected-hostiles		=expected		; expected hostiles based on quad search cycle with highest count
		< Nhostiles				=expected		; There is 1 less hostile than expected
		task					=task
		Nhostiles				=count
	?visual>
		state       			free
	?visual-location>
		state					error
==>
	=goal>
		Nhostiles				0
		ew-alarm				green 			; set ew alarm state to on
	+temporal>
		isa 					time
		ticks					0
	; If task is EW, set state to find-object and ew-check to yes - this will bypass checking ew-alarm as its silly to keep checking ew-alarm in EW only task 
	; If task is MOT/DUAL set state to search-quad and ew-check to no - this will initiate quad search 
	!eval!						(cond
								((chunk-slot-equal =task 'EW) (and (mod-buffer-chunk 'goal '(state find-object)) (mod-buffer-chunk 'goal '(ew-check yes))))
								((chunk-slot-equal =task 'MOT) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								((chunk-slot-equal =task 'DUAL) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								)
	!eval!						(remove-visual-finsts t nil)
	!output!					(=count hostiles were found but =expected were expected... the ew alarm is already on so its fine)
)

(p ew-decision-alarm-off-ok ; If the number of hostiles counted is the amount expected and the alarm is off...do nothing and end the EW check 
	=goal>
		state					attend
		task					=task
		quad					ew-all			; ew check is ongoing
		ew-check				yes				; ew check is ongoing
		ew-alarm				black			; ew alarm is off
		expected-hostiles		=expected		; expected hostiles based on quad search cycle with highest count
		>= Nhostiles			=expected		; The amount of hostiles encoded/counted is expected or greater than expected (if ew occured at start of task)
		Nhostiles				=count
	?visual>
		state       			free
	?visual-location>
		state					error
==>
	=goal>
		Nhostiles				0
		ew-alarm				black 			; set ew alarm state to off
	+temporal>
		isa 					time
		ticks					0
	; If task is EW, set state to find-object and ew-check to yes - this will bypass checking ew-alarm as its silly to keep checking ew-alarm in EW only task 
	; If task is MOT/DUAL set state to search-quad and ew-check to no - this will initiate quad search 
	!eval!						(cond
								((chunk-slot-equal =task 'EW) (and (mod-buffer-chunk 'goal '(state find-object)) (mod-buffer-chunk 'goal '(ew-check yes))))
								((chunk-slot-equal =task 'MOT) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								((chunk-slot-equal =task 'DUAL) (and (mod-buffer-chunk 'goal '(state search-quad)) (mod-buffer-chunk 'goal '(ew-check no))))
								)
	!eval!						(remove-visual-finsts t nil)
	!output!					(=count hostiles were found and =expected were expected... the ew alarm is off so its fine)
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; END PRODUCTIONS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; END MODEL DEFINITION
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
)
