import importlib
import os
import pandas as pd
import numpy as np
import seaborn as sn
import matplotlib.pyplot as plt
import time
import re
import random
import json
from scipy.interpolate import interp1d


def main():
	filename = 'results_res_filename_stem_000.csv'
	filename = os.listdir('./results')[-1]


def individual(data_dir=None,stem=None,n_windows=2):

	if data_dir is None: data_dir = os.path.join(os.getcwd(), 'results')
	if stem is None: stem = next(x[0:-4] for x in os.listdir('./results') if (not 'meta' in x) and (x[-4:]=='.csv'))

	print(f"{'#'*50}\nAnalyzing {stem}\n{'#'*50}")
	
	# Load data
	data_raw = pd.read_csv(f"{data_dir}/verbose_{stem}.csv")
	# Load metadata
	meta = pd.read_csv(f"{data_dir}/{stem}_meta.csv", header=None, names=['variable','value'], index_col='variable')
    # Add useful metadata as a column to data
	for x in meta.index:
		if x in ['task','n_friendlies','n_hostiles','enable_fatigue','fpmc']:
			data_raw.insert(0,x,meta.loc[x,'value'])

	# # Correct the quirk where fpmc has a value even if enable_fatigue is nil
	# data_raw['fatigue'] = data_raw.apply(lambda x: 0 if x['enable_fatigue']=='nil' else x['fpmc'], axis=1)



	##########################################
	# Long form with respect to quadrant
	##########################################
	# List of DataFrames we're going to merge
	frames = []
	
	# Loop through quadrants (and also EW)
	for quad in ['ne','nw','sw','se','ew']:

		# Skip if irrelevant to the task
		if meta.loc['task','value'] == 'EW' and quad != 'ew': continue
		if meta.loc['task','value'] == 'MOT' and quad == 'ew': continue

		# DF with only the relevant columns (for the current quadrant)
		df = data_raw[['time', 'task', 'n_friendlies','n_hostiles','fpmc','enable_fatigue', 'attention','attn_x','attn_y'] + data_raw.columns[data_raw.columns.map(lambda x: f'_{quad}' in x)].tolist()].copy()
		# Rename the columns
		df.columns = df.columns.str.replace(f'_{quad}','')

		# Quadrant (or ew)
		df['quad'] = quad

		# Objective (collapsing all quadrants for MOT)
		df['objective'] = 'Electronic Attack' if quad=='ew' else 'Risk'
		
		# Add CORRECT (whether the alarm matches the risk) and calculate a rolling average called ACCURACY
		df['correct'] = df['risk'] == df['alarm']
		# df['accuracy'] = df['correct'].rolling(256).mean()*100
		
		# Add LATITUDE and LONGITUDE for graphing purposes
		df['latitude'] = 'north' if quad in ['ne','nw'] else 'south'
		df['longitude'] = 'east' if quad in ['ne','se'] else 'west'
		if quad == 'ew':
			df['latitude'] = 'all'
			df['longitude'] = 'all'

		# State (hit, miss, correct rejection, false alarm)
		def state(x):
			if x['risk'] and x['alarm']: return 'Hit'
			if x['risk'] and not x['alarm']: return 'Miss'
			if not x['risk'] and not x['alarm']: return 'Correct Rejection'
			if not x['risk'] and x['alarm']: return 'False Alarm'
			return None

		df['state'] = df.apply(state, axis=1)

		# Time windows
		df['window'] = -1 + \
						np.digitize(df['time'], # we're binning the time points
									np.linspace(
												df['time'].min(), # starting with the first time point
												df['time'].max(), # ending with the last time point
												num=n_windows, # creating this many bins (default is 5)
												endpoint=False), # don't include the last time point as a bin. If you do, that time point will be a singleton in its own bin
									right=False) # each bin is [a,b) and not (a,b]


		# ########################################
		# # Has anything interesting happened?
		# ########################################
		# # There are 16 possible state changes. That's 4 possible states at the last time point (Hit, Miss, CR, FA) times 4 possible states at this new time point.
		# # We're going to loop through the rows and account for all 16

		# # Initialize this new column
		# df['event'] = pd.Series(dtype=str)
		# df.loc[0,'event'] = 'no_change'

		# # Loop through rows
		# for i in range(1,df.shape[0]):

		# 	###############################
		# 	# No change (4 state changes)
		# 	###############################
		# 	# If nothing has changed, keep things as they are with 'no_change'.
		# 	# Hit -> Hit
		# 	# Miss -> Miss
		# 	# CR -> CR
		# 	# FA -> FA
		# 	if df.at[i-1,'state']==df.at[i,'state']:
				
		# 		# Let's call this NO CHANGE
		# 		df.at[i,'event'] = 'no_change'

		# 	########################################
		# 	# No meaningful change (4 state changes)
		# 	########################################
		# 	# Some state changes are uninteresting, but should still reset the clock (i.e., the next event should be measured relative to this time point).
		# 	# We started out wrong, but...
		# 	# ...we're suddenly right for the wrong reason
		# 	# FA -> Hit
		# 	# Miss -> CR
		# 	# ...or we're still wrong, even though the risk has changed. This is extremely unlikely (you'd have to toggle an alarm at the exact wrong moment)
		# 	# FA -> Miss
		# 	# Miss -> FA

		# 	# We were wrong...
		# 	if not df.at[i-1,'correct']:
		# 		# ...and then the risk changed
		# 		if df.at[i-1,'risk'] != df.at[i,'risk']:

		# 			# Call it a NULL EVENT
		# 			df.at[i,'event'] = 'null_event'

		# 	########################################
		# 	# New Hit (2 state changes)
		# 	########################################
		# 	# We've just correctly turned the alarm ON
		# 	# Miss -> Hit
		# 	# CR -> Hit (this is unlikely - would have to activate the alarm at the exact right moment)
			
		# 	# The alarm was OFF before...
		# 	if not df.at[i-1,'alarm']: 
		# 		# ...and now it's a HIT
		# 		if df.at[i,'state']=='hit': 

		# 			# Call it a hit
		# 			df.at[i,'event'] = 'hit'

		# 	########################################
		# 	# New Correct Rejection (2 state changes)
		# 	########################################
		# 	# We've just correctly turned the alarm OFF
		# 	# FA -> CR
		# 	# Hit -> CR (this is unlikely - would have to deactivate the alarm at the exact right moment)
			
		# 	# The alarm was ON before...
		# 	if df.at[i-1,'alarm']: 
		# 		# ...and now it's a CORRECT REJECTION
		# 		if df.at[i,'state']=='cr': 

		# 			# Call it a correct rejection
		# 			df.at[i,'event'] = 'cr'

		# 	######################################
		# 	# Hot Streak Ends (4 state changes)
		# 	######################################
		# 	# We've been correct for a while (hot streak), but now we're wrong
		# 	# Hit -> Miss
		# 	# Hit -> FA
		# 	# CR -> Miss
		# 	# CR -> FA

		# 	# We were correct...
		# 	if df.at[i-1,'correct']:

		# 		# ..and now we're wrong
		# 		if not df.at[i,'correct']:

		# 			# Call it the end of a hot streak
		# 			df.at[i,'event'] = 'end_streak'

		# #####################################################
		# # How much time elapsed for that event to take place?
		# #####################################################
		# # For example, if the event was a new "hit", what was the response time?

		# # Initialize this new column
		# df['time_elapsed'] = np.nan

		# # First, get the indices for events of interest
		# event_indices = [i for i in df.index if df.loc[i,'event'] in ['null_event','hit','cr','end_streak']]

		# # The first event is a special case where the elapsed time is calculated relative to start (should always be zero)
		# df.loc[event_indices[0],'time_elapsed'] = df.loc[event_indices[0],'time'] - df.loc[0,'time']

		# # For all others, it's calculated relative to the time of the preceding event
		# df.loc[event_indices[1:],'time_elapsed'] = df.loc[event_indices[1:],'time'].values - df.loc[event_indices[0:-1],'time'].values

		# Place in the list
		frames.append(df)


	data = pd.concat(frames, axis=0, ignore_index=True)
	# data.index.set_names(['quad','sample'], inplace=True)
	# data.reset_index(inplace=True)

	# sn.kdeplot(data=data, x='attn_x',y='attn_y', fill=True, cmap='magma', thresh=0); plt.show()



	return data


def group(data_dir=None,tag=''):

	# Set the data directory
	if data_dir is None: data_dir = os.path.join(os.getcwd(), 'results')
	
	# List of individual level dataframes that will be concatenated
	data_list = []

	# A generator for looping through runs
	stems = (x[8:-4] for x in os.listdir(data_dir) if (tag in x) and (not 'meta' in x) and (x[-4:]=='.csv'))

	# Loop through those runs
	for stem in stems:
		
		# Process the current file
		try:
			indiv = individual(data_dir=data_dir,stem=stem)
		except KeyboardInterrupt:
			print('interrupted!')
			return
		except Exception as e:
			print(f'{stem} not analyzed: {e}')
			continue

		##########################
		# Roll up the data we want
		##########################

		# Initialize
		df = indiv.groupby(['objective','quad','state','n_hostiles','fpmc','enable_fatigue','window', 'task','latitude','longitude','n_friendlies'], as_index=False).agg(len)

		# Drop irrelevant columns
		df.drop(['attention','attn_x','attn_y','risk','alarm','correct'], axis=1, inplace=True)

		# Rename columns for clarity
		df.rename(columns={'time': 'occurences'}, inplace=True)

		# Add to growing list
		data_list.append(df)

	data = pd.concat(data_list, axis=0, ignore_index=True)



		

		# ##### Metadata #####

		# # Start with the metadata (but only the fields we care about)
		# new = pd.DataFrame(
		# 		pd.Series(
		# 			{k:v for k,v in metadata.items() if k in ['n_friendlies',
		# 														'n_hostiles',
		# 														'enable_fatigue',
		# 														'condition']})).T

		# ##### Add more stats #####
		
		# # Response Times for correct actions, Durations of hot streaks (that end with an incorrect action)
		# x=data[np.isfinite(data['time_elapsed'])].groupby(['event'])['time_elapsed'].mean()

		# new['rt_hit'] = x['hit'] if 'hit' in x.index else np.nan # Time to respond to RISK by turning the ALARM ON (hit)
		# new['rt_cr'] = x['cr'] if 'cr' in x.index else np.nan # Time to respond to NO RISK by turning the ALARM OFF (correct rejection)
		# new['streak'] = x['end_streak'] if 'end_streak' in x.index else np.nan # Hot streak - how long we correctly monitored RISK or lack thereof before being wrong

		# # Average accuracy across time (0 = all quadrants wrong, 100 = all quadrants correct)
		# new['accuracy'] = data.groupby('time').mean()['correct'].mean()

		# # Dwell time: proportion of time spent in each quadrant
		# counts = data['attention'].value_counts().reindex([1,2,3,4]) # how much time spent in each quadrant. Reindexing forces there to be an answer for each quadrant
		# new['dwell_NE'] = counts[1]/sum(counts)
		# new['dwell_NW'] = counts[2]/sum(counts)
		# new['dwell_SW'] = counts[3]/sum(counts)
		# new['dwell_SE'] = counts[4]/sum(counts)

		# # Add this dataset's values to the growing DataFrame
		# if 'df' in locals().keys():
		# 	df = pd.concat([df, new], axis=0, ignore_index=True, copy=False)
		# else:
		# 	df = new

	return data

def plot_group(df):
	for i,state in enumerate(np.unique(df.state)[:]):
		fg = sn.catplot(data=df[df.state==state], kind='bar',
						x='window',y='occurences',col='n_hostiles', hue='enable_fatigue', row='objective',
						sharey=False, 
						palette=[['violet','red','green','orange'][i], 'gray'],
						)
		fg.fig.subplots_adjust(top=0.9) # make room?
		fg.fig.suptitle(state)
		plt.savefig(f'{state}.jpg', dpi=300)


def plot_many(df):

	# sn.relplot(kind='line', data=df, x='n_hostiles', y='rt_hit', hue='n_friendlies', style='n_friendlies', markers=True)

	df_melt = pd.melt(df, id_vars=['n_friendlies','n_hostiles','enable_fatigue'], value_vars=['rt_hit','rt_cr','streak','accuracy'], var_name='metric', ignore_index=True)

	h = sn.relplot(kind='line', data=df_melt, x='n_hostiles', y='value', style='enable_fatigue', hue='n_friendlies', markers=True, col='metric', col_wrap=2,
					facet_kws={'sharey':False})
	h.set(xticks=np.unique(df_melt['n_hostiles']).tolist())


	plt.show()

	sn.pairplot(data=df, vars=['rt_hit','rt_cr','streak','accuracy'], hue='n_hostiles', palette='cool'); plt.show()

def plot_single(df):

	sn.relplot(kind='line', data=df, x='time', y='accuracy', row='latitude', col='longitude', hue='quad')
	sn.catplot(data=df, kind='violin', y='rt', cut=0, x='streak_type', row='latitude', col='longitude', row_order=['north','south'], col_order=['west','east'])


	# h_line = sn.lineplot(data=data, x='time', y='accuracy')
	# # h_bar = sn.histplot(data=data, x='accuracy')

	# h_streak = sn.scatterplot(data=data, x='time', y='streak_ne', hue='correct_ne')
	# h_rt = sn.violinplot(data=data, y='rt_ne', cut=0, x='streak_type_ne')
	# h_rt = sn.stripplot(data=data, y='rt_ne', x='streak_type_ne')


	plt.show()

	sn.countplot(data=data, x='correct',hue='quad'); plt.show()

if __name__ == "__main__":
    data = main()

