#####################
#####################
# mot_agi
#####################
#####################

# An ACT-R AGI version of Dr. Betsy Fox's MOT task
# Loads the model, runs the experiment once or multiple times (i.e., a parameter sweep)
# It is assumed you are using ACTR verion 7-14


################################################################
################################################################
# Dependencies
################################################################
################################################################
import actr
import importlib
import os
import pandas as pd
import numpy as np
import seaborn as sn
import matplotlib.pyplot as plt
import time
import re
import random
import json
import analyze_mot_v3 as am

################################################################
################################################################
# Global variables
################################################################
################################################################
response = None # to hold the value of any buttons pressed in the task. Initialize to empty string
sample_period = .05# how many (simulation) seconds to run the model before incrementing the AGI task

# CHANGED MODEL NAME FOR MY MODEL (AH)
    #NOTE TO CHANGE NAME TO MODEL YOU WANT OT USE!!!!!
model = {'name':'mot-model', 'filename': 'MOT-EW-model-AH2-V5.lisp'}
width = 600; height = 600 # dimensions of the experiment window

################################################################
################################################################
# Functions
################################################################
################################################################


#################
# Main
#################
# Prepare to run the MOT task

def main(): 
    #########################################################
    # Prepare the directory structure
    #########################################################
    
    # First, make sure we're in the directory containing this script
    os.chdir(os.path.split(os.path.realpath(__file__))[0])
    
    # Create the output directories if necessary
    if not os.path.exists('plots'):
        print('Creating output directory for plots')
        os.mkdir('plots')
    if not os.path.exists('results'):
        print(r'Creating output directory for results, i.e., CSV files')
        os.mkdir('results')
    
    #########################################################
    # Make a connection with ACT-R    
    #########################################################
    # If we haven't yet made a connection...
    if actr.current_connection is None:
        connect_to_actr()
        print("Giving it another few seconds to let a window handler become available")
        time.sleep(10)

    #########################################################
    # Load the MOT model
    #########################################################
    if actr.current_model()==model['name']:
        print(f"ACT-R model {model['name']} already loaded; reloading")
        actr.reload()
    else:
        load_model(os.path.join(os.getcwd(),model['filename']))


##################
# Sweep
##################
def sweep(**kwargs):
    # Conduct a parameter/condition sweep

    ##############################################
    # Prepare to run the MOT task
    ##############################################
    main()

    if kwargs == {}:
        # Comprehensive sweep
        kwargs = {
            'task_list': ['MOT','EW','DUAL'],    
            'friendlies_present_list': [True,False],
            'n_hostiles_list': [2,4,6],
            'enable_fatigue_list': ['nil'], #Removed 't' from list so it won't run with fatigue on
            'fpmc_list': [0], #[-.050,-.075,-.100,-.125, -5],
            'iterations_list': range(25),
            'time_limit': 720,
            'visible': False,
        }



    ##############################################
    # SWEEP!
    ##############################################
    # The sweep comrises nested loops.
    # The ones the require extra calculation go first (to minimize the number of times they're recalculated)
    
    # How many runs will there be, total?
    n_runs = np.prod([len(v) for k,v in kwargs.items() if '_list' in k])
    run = 1
    successes = 0
    
    # Check for some variables. Note that dict.get returns None is key doesn't exist
    visible = kwargs.get('visible')
    
    for i in kwargs['iterations_list']:


        for task in kwargs['task_list']:
            task_str = ''
            if task in ['MOT','DUAL']: task_str += 'M'
            if task in ['EW','DUAL']: task_str += 'W'

            for friendlies_present in kwargs['friendlies_present_list']:
                if friendlies_present:
                    if task_str[-1] != 'D': task_str += 'D'
                else:
                    if task_str[-1] =='D': task_str = task_str[0:-1]
                
                for n_hostiles in kwargs['n_hostiles_list']:
                    # Calculate the number of friendlies. It's dependent on the number of hostiles
                    if not friendlies_present:
                        n_friendlies = 0
                    elif n_hostiles==2:
                        n_friendlies = 5
                    elif n_hostiles==4:
                        n_friendlies = 7
                    elif n_hostiles==6:
                        n_friendlies = 10
                    else:
                        n_friendlies = 1
                
                    for fpmc in kwargs['fpmc_list']:
                        for enable_fatigue in kwargs['enable_fatigue_list']:

                            #######################
                            # Run the experiment
                            #######################
                            print(f"{'*'*40}\nAttempting run {run} of out {n_runs} total: Participant {i} of {len(kwargs['iterations_list'])}, {n_friendlies} friendlies, {n_hostiles} hostiles, fatigue {enable_fatigue}/{fpmc}")
                            try:
                                ex = Experiment(task=task, n_friendlies=n_friendlies, n_hostiles=n_hostiles, enable_fatigue=enable_fatigue, fpmc=False, time_limit=kwargs['time_limit'],
                                                visible=False, deterministic=False, save_data=True, 
                                                filename_stem=f"{i}_{n_hostiles}_{n_friendlies}_{task_str}_{time.strftime('%d_%m_%Y_%H%M%S')}",
                                                # filename_stem=f'res_{filename_stem}_{str(run).zfill(3)}',
                                                width=width, height=height)
                                ex.run()
                                print('Complete!')
                                successes += 1
                            except KeyboardInterrupt as e:
                                return
                            except Exception as e:
                                print(f'UNSUCCESSFUL')
                                print(e)

                            print(f"{successes} successes out of {run} runs attempted")
                            run += 1


############################################
# Connect_to_actr
############################################
def connect_to_actr(actr_dir=None):
    # Launch ACT-R (if necessary) and establish a connection
    # ACTR_DIR (optional): the full path to the ACT-R directory containing run-act-r.bat

    print("No active ACT-R connection. Attempting to connect.")

    # If ACT-R is running, great! All we need to do is connect
    actr.start()

    # Did that work?
    if not actr.current_connection is None: 
        print('Success - connection made!')
        return

    # If not, we need to open ACT-R
    # This might raise an exception. That's OK - it'll tell the user what to do
    print("ACT-R not running yet. Let's try opening ACT-R")
    try:
        
        if actr_dir:
            os.startfile(os.path.join(actr_dir,'run-act-r.bat'), cwd=actr_dir)
        else:
            # Assuming ACT-R 7-14 is one directory up and within folder named "ACT-R 7.14"
            os.startfile(os.path.join('..','ACT-R 7.14', 'run-act-r.bat'), cwd=os.path.join('..','ACT-R 7.14'))
    except Exception as e:
        print(e)
        print(r"Couldn't start ACT-R 7-14. Make sure that the directory containing NOT-Models also contains ACT-R_7-14")
        print(r"If that's not possible, start ACT-R 7-14 manually and re-run this script")
        raise

    # Keep trying. It takes a few seconds
    for attempt in range(5):
        # Try again to connect
        print(f'Trying to connect. Attempt #{attempt}...')
        actr.start()
        
        # if we've made the connection, great!
        if not actr.current_connection is None: 
            print('Success - connection made!')
            break

        # If not, that's OK. ACT-R takes a few seconds to open. Wait 3 seconds and try again.
        time.sleep(4) # if not, wait a few seconds and try again

    return(actr)


############################################
# Load_model
############################################
def load_model(model_file=None):
    # Load the desired model into ACT-R

    if model_file is None:
        model_file = os.path.join(os.getcwd(), 'MOT-EW-model-AH2-V4.lisp')
    try:
        actr.load_act_r_model(model_file)
        print(f"ACT-R model {model_file} loaded")
    
    except:
        print(r"If you can't load a model, it may be because ACT-R isn't running")


################################################################
################################################################
# Classes
################################################################
################################################################

##############################
# Experiment
##############################
class Experiment():
    ########################
    # Acceptable parameters:
    ########################
    # HUMAN: is a person (as opposed to an ACT-R model) performing the task?
    # TASK: multiple object tracking (MOT), electronic warfare detection (EW), or both simultaneously (dual)
    # EW_INTERVAL: EW attacks occur once in every EW_INTERVAL window, with a random onset time (at least EW_DURATION time points before the end of the window)
    # EW_DURATION: how many time points does an EW attack last? Set to 0 for no EW. 50 time points ~1 second
    # DETERMINISTIC: do we seed the random number generator, so the program runs the same every time?
    # N_FRIENDLIES, N_HOSTILES: number of friendly aircraft (i.e., disctractors) and hostile aircraft (i.e., targets)
    # VISIBLE: if True, display the window. If False, it will run faster
    # WIDTH, HEIGHT: of the window, in pixels
    # FRIENDLY_COLOR, HOSTILE_COLOR: self-explanatory
    # VERBOSE: should ACT-R generate output?
    # TODO: fatigue, speed, other ACT-R params
    # TODO: rethink EW_INTERVAL. Needs to be in simulation time
        
    ##############################
    # Initializer
    ##############################
    
    def __init__(self, n_friendlies=None,n_hostiles=None,task=None,ew_interval=None,ew_duration=None,human=None,deterministic=None,visible=False,width=None,height=None,friendly_color=None,hostile_color=None,enable_fatigue=None,fpmc=None,verbose=None,time_limit=None,save_data=False,filename_stem=None,standard_data_logging=None,verbose_data_logging=None):
        ############################
        # Set attributes
        ############################

        # Conditions of the simulation
        self.n_friendlies = n_friendlies if not n_friendlies is None else 0
        self.n_hostiles = n_hostiles if not n_hostiles is None else 1
        self.task = task if not task is None else 'DUAL'
        self.ew_interval = ew_interval if not ew_interval is None else 150
        self.ew_duration = ew_duration if not ew_duration is None else 30
        self.time_limit = time_limit if not time_limit is None else 10000
        self.human = human if not human is None else False
        self.deterministic = deterministic if not deterministic is None else True
      
        # Visual properties of the AGI window and simulation        
        self.visible = visible if not visible is None else False
        self.width = width if not width is None else 600
        self.height = height if not height is None else 600
        self.friendly_color = friendly_color if not friendly_color is None else 'red'
        self.hostile_color = hostile_color if not hostile_color is None else 'magenta'

        # Parameters for the ACT-R model
        self.enable_fatigue = enable_fatigue if not enable_fatigue is None else 'nil'
        self.fpmc = fpmc if not fpmc is None else -.05
        self.verbose = verbose if not verbose is None else False

        # Parameters for saving data to a .csv file
        self.save_data = save_data if not save_data is None else True
        self.filename_stem = filename_stem if not filename_stem is None else None
        self.standard_data_logging = standard_data_logging if not standard_data_logging is None else True
        self.verbose_data_logging = verbose_data_logging if not verbose_data_logging is None else True

        # Grab the parameters. We're going to want to save them as metadata in the output file
        self.params = locals().copy()

        # Initialize the history variable that will eventually be saved as output (if self.save_data)
        self.history = {'standard':[], 'verbose':[]}

        ########################################
        # Create the window
        ########################################
        # ...and data structures to track information about the experiment
        self.draw_window()

        ##################################
        # Prepare the ACT-R model
        ##################################
        # Reset the model
        actr.reset()
        # actr.reload()

        # Connect the model to the window
        actr.install_device(self.window)

        # Set fatigue parameters for the model
        if self.enable_fatigue=='t':
            actr.set_parameter_value(':enable-fatigue', 't')
            actr.set_parameter_value(':fpmc', fpmc)
        elif self.enable_fatigue=='nil':
            actr.set_parameter_value(':enable-fatigue', 'nil')
        else:
            print(f"enable_fatigue is set to {self.enable_fatigue}. Please use either 't' or 'nil'")    

     
        ######################################
        # Track the data, if desired 
        ######################################
        if self.save_data:

            # Make sure results directory exists
            if not os.path.isdir('results'): os.mkdir('results')
            
            # Variables of interest
            self.voi = {
                'standard': [
                                'Quadrant', # 0 = Electronic Warfare: not a quadrant, but you get the idea. 1 = North East, 2 = North West, 3 = South West, 4 = South East
                                'EventID', # 0 = mismatch begins (unsafe state), 1 = mismatch ends (safe state), 2 = alarm raised, 3 = alarm cancelled, 4 = false alarm raised, 5 = false alarm cancelled, 6 = counts changed (any change to the exp e.g., # of distractors in quad)
                                'Time', # in seconds since the start of the session
                                'Red Circle (Enemy)','Magenta Circle','Pink Circle','Red Diamond','Magenta Diamond','Pink Diamond','Red Octagon','Magenta Octagon','Pink Octagon', # how many of each shape are present
                                'Safe', # TRUE = no enemy present, FALSE = enemy present
                                'Alert', # whether the risk/MOT alarm is ON for that quadrant (undefined for Quadrant==0)
                                'Interference', # Whether an EW attack is underway
                                'Source', # where did this change come from (the experimental paradigm itself, or an action by the Subject? "exp" = (0,1,6); "Subject" = (2,3,4)
                                ],
                'verbose': [
                            'time', # simulation time, in seconds
                            'risk_ne', # is there a hostile in the northeast?
                            'risk_nw', # is there a hostile in the northwest?
                            'risk_sw', # is there a hostile in the southwest?
                            'risk_se', # is there a hostile in the southeast?
                            'alarm_ne', # is the alarm set in the northeast?
                            'alarm_nw', # is the alarm set in the northwest?
                            'alarm_sw', # is the alarm set in the southwest?
                            'alarm_se', # is the alarm set in the southeast?
                            'risk_ew', # has a track disappeared?
                            'alarm_ew', # is the EW alarm set?
                            'attention', # which quadrant is being attended?
                            'attn_x', # where exactly is attention allocated (if at all)?
                            'attn_y', # where exactly is attention allocated (if at all)?
                            ]}
                            

            # Name the output files
            if self.filename_stem is None:
                # By default, use the current time
                self.data_output_files = {
                    'standard': f'results/{time.strftime("%Y%m%d_%H%M_%S")}.csv', # YearMonthDay_HourMinute_Second
                    'verbose': f'results/{time.strftime("%Y%m%d_%H%M_%S")}_verbose.csv'
                    }
            else:
                self.data_output_files = {
                    'standard': f'results/{filename_stem}.csv',
                    'verbose': f'results/{filename_stem}_verbose.csv',
                    }

# Trying out a new method where we track the history in a variable, and only save at the end of a run
            # # Write the headers (i.e., column names, variables that change moment to moment)
            # # For standard logging
            # if self.standard_data_logging:
            #     with open(self.data_output_files['standard'], 'w') as f_data:
            #         f_data.write(','.join(self.voi['standard'])+'\n')
            # # For verbose logging
            # if self.verbose_data_logging:
            #     with open(self.data_output_files['verbose'], 'w') as f_data:
            #         f_data.write(','.join(self.voi['verbose'])+'\n')
                
            # Write the metadata (i.e., parameters for this instance of the sweep)
            # Metadata is the same for standard vs. verbose
            with open(self.data_output_files['standard'][0:-4] + '_meta' + self.data_output_files['standard'][-4:], 'w') as f_meta:
                # f_meta.write(json.dumps(self.params)+'\n')
                for k,v in self.params.items():
                    f_meta.write(k + ',' + str(v) + '\n')


    ##### Class functions for Experiment() #####


   
    ####################################################
    # Experiment.run
    ####################################################
    def run(self):
        # Run the Experiment! This is a high-level wrapper; increment_agi_task does most of the heavy lifting.

        #######################
        # Prepare commands
        #######################
        
        ##### Increment the simulation ####
        # Add the simulation incrementing function to ACT-R as a command called "interact-with-python"
        actr.add_command("interact-with-python", self.interact_with_python,
                             # "If you feel like adding a documentation string, do it here",
                             )
        # Schedule the first iteration of ACT-R calling the python function (to run the AGI task)
        actr.schedule_event_relative(0,"interact-with-python")

        ##### Button presses #####
        # Add the button pressing function to ACT-R as a command called "mot-key-press"
        actr.add_command("mot-key-press", self.respond_to_key_press,
                             # "If you feel like adding a documentation string, do it here",
                             )
        # Tell ACT-R to monitor for a key press ("output-key") and run "mot-key-press" when it detects a key press
        actr.monitor_command("output-key","mot-key-press")

        ##### Performing the right task(s) #####
        # Add the task specifying function to ACT-R as a command called "specify-task"
        actr.add_command("specify-task", self.specify_task,
                             # "If you feel like adding a documentation string, do it here",
                             )
        # Schedule the goal-setting to happen just once   
        actr.schedule_event_relative(0,"specify-task")


        ##############################
        # Run the ACT-R model
        ##############################
        # actr.current_connection.interface.show_output = 't' if verbose else 'nil'
        actr.run(self.time_limit, True)

        actr.remove_command_monitor("output-key","mot-key-press")
        actr.remove_command("mot-key-press")
        actr.remove_command("interact-with-python")
        actr.remove_command("specify-task")
        # actr.current_connection.evaluate_single("close-exp-window")
        # actr.stop()


    ####################################################
    # Experiment.draw_window
    ####################################################
    def draw_window(self):
        # Create the AGI experiment window and define background variables to track its state

        ############################
        # Draw the experiment window
        ############################
        self.window = actr.open_exp_window("MOT", visible=self.visible, width=self.width, height=self.height)

        ############################
        # Define each quadrant
        ############################
        # TOP is the y-coordinate of the northern boundary. LEFT is the x-coordinate of the western boundary. Etc.
        # CORNER_[X/Y] is the x or y coordinates of the corner furthest from the center of the screen
        # Note that there's a 2 pixel perimeter around the field (i.e., the window is 4x4 bigger than the field)
        self.quad = pd.DataFrame.from_dict(
            {'ne': {'left':width/2,
                    'right':width-2,
                    'top':2,
                    'bottom':height/2,
                    'corner_x':width-2,
                    'corner_y':2,
                    },
            'nw': {'left':2,
                    'right':width/2,
                    'top':2, 
                    'bottom':height/2,
                    'corner_x':2,
                    'corner_y':2,},
            'sw': {'left':2, 
                    'right':width/2,
                    'top':height/2,
                    'bottom':height-2, 
                    'corner_x':2,
                    'corner_y':height-2,},
            'se': {'left':width/2,
                    'right':width-2, 
                    'top':height/2,
                    'bottom':height-2,
                    'corner_x':width-2,
                    'corner_y':height-2,},
            }, orient='index', # quadrants are rows, not columns
                dtype=int) # cast as int

        # There are five pairs of lines to draw
        # ...four pairs of short black lines around the perimenter than can be turned green when a MOT alarm is triggered
        self.lines = {k:[actr.add_line_to_exp_window(self.window,
                                                self.quad.loc[k,['corner_x','corner_y']].tolist(), # from the corner
                                                [self.width/2, int(self.quad.loc[k,'corner_y'])], # to the vertical midline
                                                color='black'), # black because alarm starts OFF
                    actr.add_line_to_exp_window(self.window,
                                                self.quad.loc[k,['corner_x','corner_y']].tolist(),
                                                [int(self.quad.loc[k,'corner_x']), self.height/2], # to the horizontal midline
                                                color='black')]
                for k in self.quad.index}

        # ...one pair of long black lines (vertical and horizontal) through the center that can be turned green when the EW alarm is triggered
        self.lines['ew'] = [actr.add_line_to_exp_window(self.window, [self.width/2,2], [self.width/2,self.height-2], color='black'),
                        actr.add_line_to_exp_window(self.window, [2,self.height/2], [self.width-2,self.height/2], color='black'),
                        ]

        # Create and add to window the friendly and hostile objects. These are going to be class objects with a fair bit of info
        if self.deterministic: random.seed(0) # seed the random generator the same way every time
        self.friendlies = [Friendly(self.window) for x in range(self.n_friendlies)]
        self.hostiles = [Hostile(self.window) for x in range(self.n_hostiles)]

        # Track the current alarm status (all alarms off to begin)
        self.alarm_status = {'ne': False, # Northeast
                            'nw': False, # Northwest
                            'sw': False, # Southwest
                            'se': False, # Southeast
                            'ew': False, # Electronic Warfare
                            }

        # Track what the alarm status *should* be (all alarms off to begin)
        self.alarm_truth = {'ne': False, # Northeast
                            'nw': False, # Northwest
                            'sw': False, # Southwest
                            'se': False, # Southeast
                            'ew': False, # Electronic Warfare
                            }

        # This is how we'll track events. Initialize to everything being safe and all alarms off
        self.alarm_truth_prev = self.alarm_truth.copy()
        self.alarm_status_prev = self.alarm_status.copy()

        # Initialize the EW_WINDOW_TIME. That is, where we are in the EW_INTERVAL
        self.ew_window_time = 0



    #################################################
    # Experiment.increment_agi_task
    #################################################
    def increment_agi_task(self):
        # Run the experiment through a single iteration

        # This chunk of code is relevant only if the EW task is to be performed
        if self.task in ['EW','DUAL']:
            # If we've just started a new EW Window:
            if self.ew_window_time==0:

                # Choose a new EW_ONSET randomly. That is, when to start the EW attack.
                # Give ourselves room to complete the attack before the end of the window. We'll start the EW attack when ew_window_time == ew_onset
                self.ew_onset = random.randrange(self.ew_interval - self.ew_duration)

            ##### Electronic Warfare - Electronic Attack! #####
            if self.ew_window_time == self.ew_onset: # This is when the EW is set to occur
                x = random.choice(self.hostiles) # Pick a random hostile
                if self.task in ['EW','DUAL']:
                    # If we're doing the EW task or the DUAL task, make something disappear
                    x.invisible_time = self.ew_duration # Hide that hostile for EW_DURATION number of time points
                    # actr.remove_items_from_exp_window(window, x.id) # remove the item entirely - technique not being used
                    actr.modify_text_for_exp_window(x.id, text="ew") # instead of removing the item, change it to an "ew"
                else:
                    # If we're doing (only) the MOT task, do nothing
                    pass

        
        ##### Move all the aircraft one time step #####
        for x in self.friendlies + self.hostiles:
            # Move the aircraft in the backend
            x.pos = x.pos + (x.destination-x.pos)*x.speed/np.linalg.norm((x.destination-x.pos))
            
            # If we've reached our destination, pick a new destination
            if np.linalg.norm((x.destination-x.pos)) < x.speed:
                x.destination = np.array([self.width*random.random(), self.height*random.random()])

            # Move the aircraft in the AGI
            actr.modify_text_for_exp_window(x.id, x=x.pos[0], y=x.pos[1])

            # Hostiles that have been victoms of an EW attack
            if x in self.hostiles and x.invisible_time >= 0:
                # Take one time point off the clock
                x.invisible_time -= 1
                # If it's time for it to reappear, bring it back
                if x.invisible_time == -1:
                    actr.modify_text_for_exp_window(x.id, text=x.symbol) # change it back from a blank space to its correct symbol

        ##### Update the ground truth about the correct alarm status #####
        # i.e., risk alarms should be activated iff there is a hostile in the quadrant
        # i.e., EW alarm should be activated iff an attack has taken place
        self.alarm_truth = {q:any([self.quad.loc[q,'left'] < x.pos[0] < self.quad.loc[q,'right'] and
                                    self.quad.loc[q,'top'] < x.pos[1] < self.quad.loc[q,'bottom']
                                    for x in self.hostiles])
                            for q in self.quad.index} | \
                    {'ew': any([x.invisible_time > 0 for x in self.hostiles])} # the "|" lets you join dicts
        
        ##### Give the model or human a chance to act #####
        if self.human:
            # If it's a human, check for button presses
            actr.process_events()


        ##### If a button was pressed, do things #####
        # Six things can happen:
        # (i,r,f,j): A risk alarm is toggled for one of the four quadrants (ne,nw,sw,se)
        # (b): The EW alarm is toggled
        # (x): Exit the experiment
        global response
        
        # If no button was pressed, do nothing
        if response is None:    
            pass
        
##        # Button press for toggling an alarm - key presses
##        elif response in ['i','r','f','j','b']:
##            # Which alarm are we toggling?
##            alarm = {'i':'ne',
##                    'r': 'nw',
##                    'f': 'sw',
##                    'j': 'se',
##                    'b': 'ew'}[response]

        # Button press for toggling an alarm - using punch
        elif response in ['j','f','d','k','space']:
            # Which alarm are we toggling?
            alarm = {'j': 'ne',
                     'f': 'nw',
                     'd': 'sw',
                     'k': 'se',
                     'space': 'ew'}[response]
            
            # Toggle the status of the appropriate alarm
            self.alarm_status[alarm] = not self.alarm_status[alarm]

            # Grab the appropriate pair of lines and change the color of the lines (black for off, green for on)
            for line in self.lines[alarm]:
                actr.modify_line_for_exp_window(line, None, None, color='green' if self.alarm_status[alarm] else 'black')

            # Reset the global variable to None
            response = None

        # Write data (if requested)
        # if self.save_data: self.write_data_to_file()
        if self.save_data:
            # Old way of doing things - safe to delete if data logging is working
            # self.write_data_to_file()
            self.log_data()

        # Manual break with the "x" key
        elif response == 'x':
            # 'x' will end the experiment
            self.end_experiment()
            # response = None
            # keep_running = False
            # break
        
        # Time limit reached
        if actr.get_time()/1000 >= self.time_limit:
            self.end_experiment()     

        # Save a copy of the current status as "prev" for the next increment
        self.alarm_truth_prev = self.alarm_truth.copy()
        self.alarm_status_prev = self.alarm_status.copy()

        # If we've just finished an EW Window, reset EW_WINDOW_TIME
        if self.ew_window_time==self.ew_duration:
            self.ew_window_time == 0



    #################################################
    # Experiment.write_data_to_file
    #################################################
    def write_data_to_file(self):
        # Collect and save data if requested. There are two modes of data logging:
        # Standard data logging creates a file that is identical in format to those produced by AFRL's tkinter-based verion of this task, designed for human participants. Each row is an event of interest.
        # Verbose data logging saves the state of the experiment every <sample_period> seconds, regardless of whether anything interesting has happened
        # The user can choose either, both, or neither mode of data logging

        # Verbose data-logging (i.e., recording the state of the experiment every iteration)
        if self.verbose_data_logging:
            # Initialize single-row data frame
            # Make sure the variables of interest are correctly lined up in the output file
            # See the first few lines in the "Run the experiment!" section where the voi are defined, around line 221
            data = pd.DataFrame(columns=self.voi['verbose'], index=[0], dtype=object)

            data['time'] = actr.get_time()/1000 # simulation time, in seconds
            data['risk_ne'] =  self.alarm_truth['ne'] # is there a hostile in the northeast?
            data['risk_nw'] =  self.alarm_truth['nw'] # is there a hostile in the northwest?
            data['risk_sw'] =  self.alarm_truth['sw'] # is there a hostile in the southwest?
            data['risk_se'] =  self.alarm_truth['se'] # is there a hostile in the southeast?
            data['alarm_ne'] =  self.alarm_status['ne'] # is the alarm set in the northeast?
            data['alarm_nw'] =  self.alarm_status['nw'] # is the alarm set in the northwest?
            data['alarm_sw'] =  self.alarm_status['sw'] # is the alarm set in the southwest?
            data['alarm_se'] =  self.alarm_status['se'] # is the alarm set in the southeast?
            data['risk_ew'] =  self.alarm_truth['ew'] # is an EW attack underway (i.e., a vanished track)?
            data['alarm_ew'] =  self.alarm_status['ew'] # is the EW alarm set?
            data['attention'] = actr.chunk_slot_value(actr.buffer_chunk('goal')[0], 'quadnum') # which quadrant is being attended? NE=1, NW=2, SW=3, SE=4
            if actr.buffer_chunk('visual-location')[0]:
                # If there's nothing in visual-location, the next two lines give an annoying warning
                data['attn_x'] = actr.chunk_slot_value(actr.buffer_chunk('visual-location')[0], 'screen-x') # exactly where are we looking (if anywhere)?
                data['attn_y'] = actr.chunk_slot_value(actr.buffer_chunk('visual-location')[0], 'screen-y')
                
                ### Get the utilities in real time for the various productions?
                # TODO

            ##### Write data to file #####
            with open(self.data_output_files['verbose'], 'a') as f_data:
                data.to_csv(f_data, header=False, index=False, lineterminator='\n')

                ######################################################################

        if self.standard_data_logging:
            # Standard data logging
            data = pd.DataFrame(columns=self.voi['standard'], dtype=object)

            # Instantiate the full list of these that will be concatenated
            data_list = []

            # We'll be making sure the variables of interest are correctly lined up in the output file by specifying index=self.voi
            # See the first few lines in the "Run the experiment!" section where the voi are defined, around line 221

            # Loop through quadrants (including EW)
            for quad_num,quad_name in ((0,'ew'),(1,'ne'),(2,'nw'),(3,'sw'),(4,'se')):

                ##### Did anything happen in this quadrant? #####
                # Note that two events here (one for the actual risk, one for participant actions) are possible at the same time

                ##### Events #0 and #1 are abut the risk state changing
                if not self.alarm_truth_prev[quad_name] and self.alarm_truth[quad_name]:
                    # EventID #0: new threat
                    data_list.append(pd.Series({'Quadrant':quad_num,
                                                'EventID':0,
                                                'Safe':False,
                                                'Alert':self.alarm_status[quad_name],
                                                'Source':'exp'}, index=self.voi['standard']))
                elif self.alarm_truth_prev[quad_name] and not self.alarm_truth[quad_name]:
                    # EventID #1: threat ends, safe state begins
                    data_list.append(pd.Series({'Quadrant':quad_num,
                                                'EventID':1,
                                                'Safe':True,
                                                'Alert':self.alarm_status[quad_name],
                                                'Source':'exp'}, index=self.voi['standard']))
            
                ##### Events #2 and #4 are about alarms being raised by the participant/subject
                if not self.alarm_status_prev[quad_name] and self.alarm_status[quad_name]:                      
                    if self.alarm_truth[quad_name]:
                        # EventID #2: alarm correctly raised (hit)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':2,
                                                    'Safe':False,
                                                    'Alert':True,
                                                    'Source':'Subject'}, index=self.voi['standard']))
                    else:
                        # EventID #4: alarm incorrectly raised (false alarm)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':4,
                                                    'Safe':True,
                                                    'Alert':True,
                                                    'Source':'Subject'}, index=self.voi['standard']))                    
                ##### Events #3 and #5 are about alarms being cancelled by the participant              
                elif self.alarm_status_prev[quad_name] and not self.alarm_status[quad_name]:
                    if not self.alarm_truth[quad_name]:
                        # EventID #3: alarm correctly cancelled (correct rejection)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':3,
                                                    'Safe':True,
                                                    'Alert':False,
                                                    'Source':'Subject'}, index=self.voi['standard']))
                    else:
                        # EventID #5: alarm incorrectly cancelled (miss)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':5,
                                                    'Safe':False,
                                                    'Alert':False,
                                                    'Source':'Subject'}, index=self.voi['standard']))

            # Write data only if something has happened
            if len(data_list) == 0:

                # Is nothing has happened, we won't have collected any data. Pass the rest of the THEN clause
                pass

            else:
                # Concatenate all quadrants' information into one DataFrame
                data = pd.concat(data_list, axis=1).T

                ##### Fill in the rest of the columns. Some of this information is redundant ¯\_(ツ)_/¯
                data['Time'] = actr.get_time()/1000 # simulation time, in seconds
                data.loc[data['Quadrant']==0,'Red Circle (Enemy)'] = ''
                data.loc[data['Quadrant']!=0,'Red Circle (Enemy)'] = sum([(self.quad.loc[quad_name,'left'] < x.pos[0] < self.quad.loc[quad_name,'right']) and
                                                                            (self.quad.loc[quad_name,'top'] < x.pos[1] < self.quad.loc[quad_name,'bottom'])
                                                                            for x in self.hostiles])
                data.loc[data['Quadrant']==0,'Magenta Diamond'] = ''
                data.loc[data['Quadrant']!=0,'Magenta Diamond'] = sum([(self.quad.loc[quad_name,'left'] < x.pos[0] < self.quad.loc[quad_name,'right']) and
                                                                        (self.quad.loc[quad_name,'top'] < x.pos[1] < self.quad.loc[quad_name,'bottom'])
                                                                        for x in self.friendlies])
                data[['Magenta Circle',
                        'Pink Circle',
                        'Red Diamond',
                        'Pink Diamond',
                        'Red Octagon',
                        'Magenta Octagon',
                        'Pink Octagon',
                        ]] = 0
                data['Interference'] = self.alarm_truth['ew']

                ##### Write data to file #####
                with open(self.data_output_files['standard'], 'a') as f_data:
                    data.to_csv(f_data, header=False, index=False, lineterminator='\n')
                # data_list.append(data)


    #################################################
    # Experiment.log_data
    #################################################
    def log_data(self):
        # Collect and data if requested, to be saved at the end of the run. There are two modes of data logging:
        # Standard data logging creates a file that is identical in format to those produced by AFRL's tkinter-based verion of this task, designed for human participants. Each row is an event of interest.
        # Verbose data logging saves the state of the experiment every <sample_period> seconds, regardless of whether anything interesting has happened
        # The user can choose either, both, or neither mode of data logging

        # Verbose data-logging (i.e., recording the state of the experiment every iteration)
        if self.verbose_data_logging:
            # Initialize single-row data frame
            # Make sure the variables of interest are correctly lined up in the output file
            # See the first few lines in the "Run the experiment!" section where the voi are defined, around line 221
            data = pd.DataFrame(columns=self.voi['verbose'], index=[0], dtype=object)

            data['time'] = actr.get_time()/1000 # simulation time, in seconds
            data['risk_ne'] =  self.alarm_truth['ne'] # is there a hostile in the northeast?
            data['risk_nw'] =  self.alarm_truth['nw'] # is there a hostile in the northwest?
            data['risk_sw'] =  self.alarm_truth['sw'] # is there a hostile in the southwest?
            data['risk_se'] =  self.alarm_truth['se'] # is there a hostile in the southeast?
            data['alarm_ne'] =  self.alarm_status['ne'] # is the alarm set in the northeast?
            data['alarm_nw'] =  self.alarm_status['nw'] # is the alarm set in the northwest?
            data['alarm_sw'] =  self.alarm_status['sw'] # is the alarm set in the southwest?
            data['alarm_se'] =  self.alarm_status['se'] # is the alarm set in the southeast?
            data['risk_ew'] =  self.alarm_truth['ew'] # is an EW attack underway (i.e., a vanished track)?
            data['alarm_ew'] =  self.alarm_status['ew'] # is the EW alarm set?
            data['attention'] = actr.chunk_slot_value(actr.buffer_chunk('goal')[0], 'quadnum') # which quadrant is being attended? NE=1, NW=2, SW=3, SE=4
            if actr.buffer_chunk('visual-location')[0]:
                # If there's nothing in visual-location, the next two lines give an annoying warning
                data['attn_x'] = actr.chunk_slot_value(actr.buffer_chunk('visual-location')[0], 'screen-x') # exactly where are we looking (if anywhere)?
                data['attn_y'] = actr.chunk_slot_value(actr.buffer_chunk('visual-location')[0], 'screen-y')
                
                ### Get the utilities in real time for the various productions?
                # TODO

            ##### Store data in HISTORY #####
            self.history['verbose'].append(data)


                ######################################################################

        if self.standard_data_logging:
            # Standard data logging
            data = pd.DataFrame(columns=self.voi['standard'], dtype=object)

            # Instantiate the full list of these that will be concatenated
            data_list = []

            # We'll be making sure the variables of interest are correctly lined up in the output file by specifying index=self.voi
            # See the first few lines in the "Run the experiment!" section where the voi are defined, around line 221

            # Loop through quadrants (including EW)
            for quad_num,quad_name in ((0,'ew'),(1,'ne'),(2,'nw'),(3,'sw'),(4,'se')):

                ##### Did anything happen in this quadrant? #####
                # Note that two events here (one for the actual risk, one for participant actions) are possible at the same time

                ##### Events #0 and #1 are abut the risk state changing
                if not self.alarm_truth_prev[quad_name] and self.alarm_truth[quad_name]:
                    # EventID #0: new threat
                    data_list.append(pd.Series({'Quadrant':quad_num,
                                                'EventID':0,
                                                'Safe':False,
                                                'Alert':self.alarm_status[quad_name],
                                                'Source':'exp'}, index=self.voi['standard']))
                elif self.alarm_truth_prev[quad_name] and not self.alarm_truth[quad_name]:
                    # EventID #1: threat ends, safe state begins
                    data_list.append(pd.Series({'Quadrant':quad_num,
                                                'EventID':1,
                                                'Safe':True,
                                                'Alert':self.alarm_status[quad_name],
                                                'Source':'exp'}, index=self.voi['standard']))
            
                ##### Events #2 and #4 are about alarms being raised by the participant/subject
                if not self.alarm_status_prev[quad_name] and self.alarm_status[quad_name]:                      
                    if self.alarm_truth[quad_name]:
                        # EventID #2: alarm correctly raised (hit)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':2,
                                                    'Safe':False,
                                                    'Alert':True,
                                                    'Source':'Subject'}, index=self.voi['standard']))
                    else:
                        # EventID #4: alarm incorrectly raised (false alarm)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':4,
                                                    'Safe':True,
                                                    'Alert':True,
                                                    'Source':'Subject'}, index=self.voi['standard']))                    
                ##### Events #3 and #5 are about alarms being cancelled by the participant              
                elif self.alarm_status_prev[quad_name] and not self.alarm_status[quad_name]:
                    if not self.alarm_truth[quad_name]:
                        # EventID #3: alarm correctly cancelled (correct rejection)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':3,
                                                    'Safe':True,
                                                    'Alert':False,
                                                    'Source':'Subject'}, index=self.voi['standard']))
                    else:
                        # EventID #5: alarm incorrectly cancelled (miss)
                        data_list.append(pd.Series({'Quadrant':quad_num,
                                                    'EventID':5,
                                                    'Safe':False,
                                                    'Alert':False,
                                                    'Source':'Subject'}, index=self.voi['standard']))

            # Write data only if something has happened
            if len(data_list) == 0:

                # Is nothing has happened, we won't have collected any data. Pass the rest of the THEN clause
                pass

            else:
                # Concatenate all quadrants' information into one DataFrame
                data = pd.concat(data_list, axis=1).T

                ##### Fill in the rest of the columns. Some of this information is redundant ¯\_(ツ)_/¯
                data['Time'] = actr.get_time()/1000 # simulation time, in seconds
                data.loc[data['Quadrant']==0,'Red Circle (Enemy)'] = ''
                data.loc[data['Quadrant']!=0,'Red Circle (Enemy)'] = sum([(self.quad.loc[quad_name,'left'] < x.pos[0] < self.quad.loc[quad_name,'right']) and
                                                                            (self.quad.loc[quad_name,'top'] < x.pos[1] < self.quad.loc[quad_name,'bottom'])
                                                                            for x in self.hostiles])
                data.loc[data['Quadrant']==0,'Magenta Diamond'] = ''
                data.loc[data['Quadrant']!=0,'Magenta Diamond'] = sum([(self.quad.loc[quad_name,'left'] < x.pos[0] < self.quad.loc[quad_name,'right']) and
                                                                        (self.quad.loc[quad_name,'top'] < x.pos[1] < self.quad.loc[quad_name,'bottom'])
                                                                        for x in self.friendlies])
                data[['Magenta Circle',
                        'Pink Circle',
                        'Red Diamond',
                        'Pink Diamond',
                        'Red Octagon',
                        'Magenta Octagon',
                        'Pink Octagon',
                        ]] = 0
                data['Interference'] = self.alarm_truth['ew']

                ##### Store data in HISTORY #####
                self.history['standard'].append(data)

    def end_experiment(self):
        # This is not currently in use, but may be useful later in development        
        # #############################################
        # # Clean up
        # #############################################
        # actr.remove_command_monitor("output-key","mot-key-press")
        # actr.remove_command("mot-key-press")
        # actr.remove_command("interact-with-python")
        # actr.current_connection.evaluate_single("close-exp-window")
        # actr.stop()

        if self.save_data:
            if self.standard_data_logging:
                df = pd.concat(self.history['standard'], axis=0, ignore_index=True)
                df.to_csv(self.data_output_files['standard'], header=True, index=False, mode='w',lineterminator='\n')
            if self.verbose_data_logging:
                df = pd.concat(self.history['verbose'], axis=0, ignore_index=True)
                df.to_csv(self.data_output_files['verbose'], header=True, index=False, mode='w',lineterminator='\n')



    ####################################################
    # Send ACT-R commands to control timing, key press
    ####################################################
    # Note that this should work for a human or model. Untested for humans


    ####################################################
    # Experiment.interact_with_python
    
    def interact_with_python(self):
        # Manage timing.
        # This short yet vital function is sent to ACT-R as a command. Every <sample_period> seconds, it will be run, incrementing the simulation by a single time point
        # Experiment.increment_agi_task() does the heavy lifting

        global sample_period

        self.increment_agi_task()
        
        # Schedule the next iteration to run the AGI task
        actr.schedule_event_relative(sample_period,"interact-with-python")
        # TODO: sample period should be equal to the number of real seconds between screen updates in betsy's mot

    ####################################################
    # Experiment.response_to_key_press
    
    def respond_to_key_press (self, model, key): 
        # Manage key presses. When the model "presses" a key via its manual module, it will be registered in the global "response" variable.
        # This function is also sent to ACT-R as a command, to be triggered whenever a button is pressed

        # Access the global variable RESPONSE that is defined at the top of this script
        global response
        # Assign to it the value of the key that was pressed
        response = key

    ####################################################
    # Experiment.specify_task
    
    def specify_task(self):
        # Set the task slot in the goal buffer with one of {MOT, EW, DUAL}

        actr.set_chunk_slot_value(
            actr.buffer_chunk('goal')[0], 
            'task',
            self.task)





if __name__ == "__main__":
    main()


################################################################################
# Aircraft classes
################################################################################

class Aircraft():
    # Parent class for all aircraft types. Never used except for inheritance

    def __init__(self, window, window_size=None, symbol=None, color=None, invisible_time=0, pos=None, destination=None, speed=None, size=None):

        ###############
        # Parameters
        ###############
        # WINDOW: an active ACT-R AGI window
        # WINDOW_SIZE: dimensions, in pixels from (left,top), of the AGI window
        # SYMBOL: the symbol, usually a character, that represents the aircraft on the screen
        # COLOR: color of the aircraft. Will usually be determined in the specific class ('red' for FRIENDLY or 'magenta' for HOSTILE)
        # INVISIBLE_TIME: how much longer will it be invisible (due to EW attack)? If 0, then it's simply visible
        # POS: starting position in pixels from (left, top)
        # DESTINATION: where is it going?
        # SPEED: how many pixels does it move per time step?
        # SIZE: font size of the aircraft (default = 12)

        ##############
        # Attributes
        ##############      
        self.window_size = (600,600) if window_size is None else tuple(window_size) # make sure it's a tuple
        self.symbol = 'X' if symbol is None else str(symbol) # it shouldn't be an X. It should be specified in inherited class
        self.color = 'yellow' if color is None else str(color) # same
        self.invisible_time = 0 if invisible_time is None else int(invisible_time)
        self.pos = np.array([random.randint(0, window_size[0]), random.randint(0,window_size[1])]) if pos is None else np.array(pos) # random point within the window
        self.destination = np.array([window_size[0]*random.random(), window_size[1]*random.random()]) if destination is None else np.array(destination)
        self.speed = random.uniform(1,2) if speed is None else speed
        self.size = 100 if size is None else size
        self.id = actr.current_connection.evaluate_single("create-text-for-exp-window", window, symbol, [["color", self.color], ["font-size", self.size]]) # Create the AGI object and save its ID
        # raise
        # Add it to the window
        actr.add_items_to_exp_window(window, # add to window...
            actr.modify_text_for_exp_window(self.id, x=int(self.pos[0]), y=int(self.pos[1]))) # ... after setting the correct position


class Friendly(Aircraft):
    # Friendly aircraft

    def __init__(self, window, window_size=(600,600), symbol='F', color='red', invisible_time=0, pos=None, destination=None, speed=None, size=12):

        ###############
        # Parameters
        ###############
        # See parent class

        # Global variables (that might be the wrong term for this)
        min_speed = 2
        max_speed = 3

        ##############
        # Attributes
        ##############
        # Speed needs to be determined here
        speed = random.uniform(min_speed, max_speed) if speed is None else speed

        # Most attributes are inherited from the Aircraft superclass
        super(Friendly, self).__init__(window=window,
                                        window_size=window_size,
                                        symbol=symbol,
                                        color=color,
                                        invisible_time=invisible_time,
                                        pos=pos,
                                        destination=destination,
                                        speed=speed,
                                        size=size,
                                        )


class Hostile(Aircraft):
    # Hostile aircraft

    def __init__(self, window, window_size=(600,600), symbol='H', color='magenta', invisible_time=0, pos=None, destination=None, speed=None, size=20):

        ###############
        # Parameters
        ###############
        # See parent class

        # Global variables (that might be the wrong term for this)
        min_speed = 3
        max_speed = 5

        ##############
        # Attributes
        ##############
        # Speed needs to be determined here
        speed = random.uniform(min_speed, max_speed) if speed is None else speed

        # Most attributes are inherited from the Aircraft superclass
        super(Hostile, self).__init__(window=window,
                                        window_size=window_size,
                                        symbol=symbol,
                                        color=color,
                                        invisible_time=invisible_time,
                                        pos=pos,
                                        destination=destination,
                                        speed=speed,
                                        size=size,
                                        )
